/*
 * Copyright (C) 2014 Erik-Oliver Blass
 *                    Travis Mayberry
 *                    Guevara Noubir
 *                    Kaan Onarlioglu <onarliog@ccs.neu.edu>
 *
 * This program is free software: you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program. If not, see <http://www.gnu.org/licenses/>.
 */

/*
 * hive.h, ver.2014.11.03
 *
 * HiVE core.
 */
#ifndef HIVE_H
#define HIVE_H

#include <libdevmapper.h> /* Devmapper ioctl interface. */
#include <termios.h>      /* To disable terminal echoing for password prompt. */
#include <linux/fs.h>

#include "utils.h"

#define HIVE_TARGET "hive"
#define HIVE_MAP_DEV_PREFIX "map-"
#define HIVE_PUBLIC_DEV_SUFFIX "_p"
#define HIVE_HIDDEN_DEV_SUFFIX "_h"
#define HIVE_DEV_PATH "/dev/mapper/"
#define HIVE_NO_MAP_HIVE ""

#define HIVE_HEADER_LEN 1     /* In sectors. */
#define HIVE_STASH_LEN 64     /* In sectors. */
#define HIVE_STASH_MAP_LEN 1  /* In sectors. */
#define HIVE_KEY_LEN 32       /* In bytes. */
#define HIVE_IV_LEN 16        /* In bytes. */

#define HIVE_SALT_LEN 32
#define HIVE_DIGEST_LEN 32

#define HIVE_MAX_WORLDS 2
#define WORLD_0 0
#define WORLD_1 1

#define HIVE_CREATE 1
#define HIVE_OPEN 2

#define HIVE_PROC_FILE "/proc/hivetab"

/* Encryption keys. */
char hive_keys[HIVE_MAX_WORLDS][HIVE_KEY_LEN];

/* No of worlds. */
int hive_worlds;

/*
 * HiVE header entry, one for each world. Some fields are not very useful at
 * the moment but that might change in the future.
 */
struct hive_header_entry {
    char key[HIVE_KEY_LEN];           /* Encrypted encryption key. */
    char key_digest[HIVE_DIGEST_LEN]; /* Key digest. */
    char key_salt[HIVE_SALT_LEN];     /* Encryption key salt. */
    char pass_salt[HIVE_SALT_LEN];    /* Password salt. */
    u64 stash_start;             /* Disk sector stash. */
    u64 stash_len;
    u64 stash_map_start;         /* Stash map. */
    u64 stash_map_len;
    u64 trans_start;             /* Sector translation map. */
    u64 trans_len;
    u64 meta_start;              /* IVs and reverse map entries. */
    u64 data_start;              /* Actual data starts here. */
    u64 data_len;                /* All sectors that can store data. */
    u64 world_len;               /* Number of data sectors in this world. */
    u64 next_hive;               /* The next recursive HiVE. */
    char iv[HIVE_IV_LEN];        /* IV for stash and translation map encryption. */
    u64 padding1;
    u64 padding2;
    u64 padding3; /* Must extend header sector size if more data needed! */
};

/*
 * IVs and reverse map entries for each data sector. Must be a multiple of AES
 * block size to avoid the padding.
 */
struct hive_meta_block {
    char iv[HIVE_IV_LEN];
    u64 sector;
    u64 world;
};

/* Size calculation helpers. */
u64 get_data_sectors(u64);
u64 get_meta_sectors(u64);
u64 get_map_sectors(u64);
u64 get_stash_sectors();
u64 get_data_hive_len(u64);
u64 get_map_hive_len(u64);
unsigned get_size_correction(u64, u64, u64);

/* Key derivation and management. */
void get_keys(int, struct hive_header_entry *);
int verify_keys(struct hive_header_entry *);
void cleanup_keys();

/* Device name derivation helper. */
void get_names(char *, char *, char **, char **, char **, char **, char **, char **, char **);

/* Map & meta block functions. */
void encrypt_and_write_sectors(int, char *, unsigned, char *, char *);
void write_map_to_other_hive(char *, u64, u64);
void write_map_to_this_hive(int, u64, u64, char *, char *);
void encrypt_meta(struct hive_meta_block *, char *);
void write_meta(int, u64, u64, u64, char *);
void write_random_meta(int, u64);

/* Actual commands. */
int close_hive(char *);
void _do_close(char **, int);
void do_close(char *);
int open_hive(char *, char *, char *, char *, u64, u64, int, char *);
int open_map_hive(char *, char *, char *, char *,  u64, u64, int);
int open_data_hive(char *, char *, char *, char *, u64, u64, int, char *);
void do_open(char *, char *, char *, char *, int);
void create_hive(int, u64, u64, char *, char *, char *);
void do_create(char *, char *, char *);
void do_list();

#endif /* HIVE_H */
